# Generated by Django 2.1.8 on 2019-10-06 04:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0010_checkoutstore_payment_received'),
    ]

    operations = [
        migrations.AlterField(
            model_name='checkoutstore',
            name='last_updated',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 2.1.8 on 2019-08-27 01:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0002_auto_20190813_1349'),
    ]

    operations = [
        migrations.AddField(
            model_name='productvariantrating',
            name='active_flag',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='productvariantrating',
            name='review',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariantrating',
            name='review_qc_flag',
            field=models.BooleanField(default=False),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.http import Http404
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView


class SaleboxAccountHistoryListView(TemplateView):
    template_name = 'salebox/account/history_list.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        context['user'] = request.user
        context['member'] = request.user.get_member()
        context['member'].transactionhistory_get_data()

        # filter data depenent on search
        s = request.GET.get('filter', '').lower()
        if len(s) > 0:
            context['member'].salebox_transactionhistory_data = [
                d for d in context['member'].salebox_transactionhistory_data if s in d['pos_guid'].lower()
            ]

        return self.render_to_response(context)


class SaleboxAccountHistoryDetailView(TemplateView):
    template_name = 'salebox/account/history_detail.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        context['user'] = request.user

        try:
            member = context['user'].get_member()
            context['member'] = member
            context['member'].transactionhistory_get_data()
            context['transaction'] = member.transactionhistory_get_single_transaction(self.kwargs['id'])
        except:
            raise Http404

        # check transaction exists
        if context['transaction'] is None:
            raise Http404

        return self.render_to_response(context)


from django import forms

from saleboxdjango.lib.shipping_options import SaleboxShippingOptions
from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutShippingMethodForm(forms.Form):
    shipping_method = forms.IntegerField()


class SaleboxCheckoutShippingMethodView(SaleboxCheckoutBaseView):
    shipping_options_class = SaleboxShippingOptions
    checkout_step = 'shipping_method'
    form_class = SaleboxCheckoutShippingMethodForm
    template_name = 'salebox/checkout/shipping_method.html'

    def get_additional_context_data(self, context):
        context = self._get_shipping_options(context)

        # print(context['shipping_options'])

        combined_price = None

        # use selected option if exists
        if self.sc.get_raw_data()['shipping_method']['id'] is not None:
            for o in context['shipping_options']:
                if o['id'] == self.sc.get_raw_data()['shipping_method']['id']:
                    combined_price = o['combined_price']
                    break

        # no option selected, choice the default
        if combined_price is None:
            for o in context['shipping_options']:
                if o['available'] and o['selected']:
                    combined_price = o['combined_price']
                    break

        # no option selected, no default, fallback to first available
        if combined_price is None:
            for o in context['shipping_options']:
                if o['available']:
                    combined_price = o['combined_price']
                    break

        context['combined_shipping_price'] = combined_price
        return context

    def form_valid_pre_redirect(self, form):
        # retrieve selected option
        id = form.cleaned_data['shipping_method']
        options = self._get_shipping_options()['shipping_options']
        selected = next(o for o in options if o['id'] == id)

        # store in checkout
        self.sc.set_shipping_method(
            selected['id'],
            selected['code'],
            selected['label'],
            selected['price'],
            selected['meta'],
            self.request
        )

        return True

    def _get_shipping_options(self, context={}):
        smc = self.shipping_options_class()
        return smc.go(
            self.request,
            self.sc.get_raw_data(),
            context
        )



# Generated by Django 2.1.8 on 2019-10-05 02:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0005_productvariant_preorder_flag'),
    ]

    operations = [
        migrations.CreateModel(
            name='Translation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language_code', models.CharField(max_length=7)),
                ('key', models.CharField(max_length=255)),
                ('value', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'Translation',
                'ordering': ['language_code', 'key'],
            },
        ),
    ]



# Generated by Django 2.1.8 on 2019-09-24 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0004_auto_20190829_1407'),
    ]

    operations = [
        migrations.AddField(
            model_name='productvariant',
            name='preorder_flag',
            field=models.BooleanField(default=False),
        ),
    ]



from django.db.models import Avg, Sum
from django.http import JsonResponse

from saleboxdjango.models import Product, ProductVariant, \
    ProductVariantRating
from saleboxdjango.lib.common import get_rating_display


class SaleboxRating:
    def __init__(self, user=None):
        self.user = user
        self.variant = None  # set via the set_variant_from_id method below

    # used by SaleboxRatingAddView / SaleboxRatingRemoveView to provide rating
    # scores for the variant just modified
    def get_data(self, results):
        results = results.split(',')
        o = {}

        if 'global_product_rating' in results:
            o['global_product_rating'] = self.get_global_product_rating()

        if 'global_variant_rating' in results:
            o['global_variant_rating'] = self.get_global_variant_rating()

        if 'user_variant_rating' in results:
            o['user_rating'] = self.get_user_variant_rating()

        return o

    # get the score for a single product (votes from all users)
    def get_global_product_rating(self):
        if self.variant is None:
            raise ValueError('SaleboxRating product variant not set')

        p = Product.objects.get(id=self.variant.product.id)
        return self._return_rating(p.rating_vote_count, p.rating_score)

    # get reviews (ratings with review text) for a single product (votes from all users)
    # TODO: pagination
    def get_global_product_reviews(self, order='-created'):
        if self.variant is None:
            raise ValueError('SaleboxRating product variant not set')

        # get variant IDs
        variant_ids = ProductVariant \
                        .objects \
                        .filter(product=self.variant.product) \
                        .values_list('id', flat=True)

        # return all reviews for all variants in above list
        return ProductVariantRating \
                .objects \
                .filter(variant=variant_ids) \
                .filter(review__inull=False) \
                .filter(active_flag=True) \
                .filter(review_qc_flag=False) \
                .select_related('user') \
                .order_by(order)

    # get the score for a single variant (votes from all users)
    def get_global_variant_rating(self):
        if self.variant is None:
            raise ValueError('SaleboxRating product variant not set')

        pv = ProductVariant.objects.get(id=self.variant.id)
        return self._return_rating(pv.rating_vote_count, pv.rating_score)

    # get reviews (ratings with review text) for a single variant (votes from all users)
    # TODO: pagination
    def get_global_variant_reviews(self, order='-created'):
        if self.variant is None:
            raise ValueError('SaleboxRating product variant not set')

        return ProductVariantRating \
                .objects \
                .filter(variant=self.variant.id) \
                .filter(active_flag=True) \
                .filter(review__isnull=False) \
                .filter(review_qc_flag=False) \
                .select_related('user') \
                .order_by(order)

    # get the score for a single variant (current user's vote only)
    def get_user_variant_rating(self):
        if self.user is not None:
            if self.variant is None:
                raise ValueError('SaleboxRating product variant not set')

            pvr = ProductVariantRating \
                        .objects \
                        .filter(variant=self.variant) \
                        .filter(user=self.user) \
                        .first()

            if pvr is None:
                return self._return_rating(0, 0)
            else:
                return self._return_rating(1, pvr.rating)

        return None

    # add (or update an existing) variant rating
    def rating_add(self, rating, review=None):
        if self.user is not None:
            if self.variant is None:
                raise ValueError('SaleboxRating product variant not set')

            pvs = ProductVariantRating \
                    .objects \
                    .filter(variant=self.variant) \
                    .filter(user=self.user)

            if len(pvs) > 0:
                pvs[0].rating = rating
                pvs[0].review = review
                pvs[0].save()
                for pv in pvs[1:]:
                    pv.delete()
            else:
                pv = ProductVariantRating(
                    user=self.user,
                    variant=self.variant,
                    rating=rating,
                    review=review
                )
                pv.save()

    # remove a variant rating
    def rating_remove(self):
        if self.user is not None:
            if self.variant is None:
                raise ValueError('SaleboxRating product variant not set')

            pvs = ProductVariantRating \
                    .objects \
                    .filter(variant=self.variant) \
                    .filter(user=self.user)

            for pv in pvs:
                pv.delete()

    # fetch the applicable variant
    def set_variant_from_id(self, variant_id):
        self.variant_id = variant_id
        self.variant = ProductVariant \
                        .objects \
                        .select_related('product') \
                        .get(id=self.variant_id)

    # private function to return results as a dict
    def _return_rating(self, count, rating):
        return {
            'count': count,
            'rating': get_rating_display(rating, count)
        }


from django import forms
from .base import SaleboxCheckoutBaseView


class SaleboxCheckoutConfirmationForm(forms.Form):
    pass


class SaleboxCheckoutConfirmationView(SaleboxCheckoutBaseView):
    checkout_step = 'confirmation'
    form_class = SaleboxCheckoutConfirmationForm
    template_name = 'salebox/checkout/confirmation.html'


from saleboxdjango.forms import SaleboxRatingAddForm
from saleboxdjango.views.rating.base import SaleboxRatingView


class SaleboxRatingAddView(SaleboxRatingView):
    action = 'rating-add'
    form = SaleboxRatingAddForm

    def form_valid(self, request):
        try:
            self.sr.set_variant_from_id(self.form.cleaned_data['variant_id'])
            self.sr.rating_add(self.form.cleaned_data['rating'])

            self.status = 'success'
            self.results = self.sr.get_data(self.results_csv)
        except:
            self.status = 'fail'



from saleboxdjango.views.base import SaleboxBaseView
from saleboxdjango.lib.rating import SaleboxRating


class SaleboxRatingView(SaleboxBaseView):
    def init_class(self, request):
        self.sr = SaleboxRating(request.user)



from django.db.models import Case, BooleanField, Value, When

from saleboxdjango.models import ProductVariant


def get_sibling_variants(variant, order_by=None):
    pv = ProductVariant \
            .objects \
            .filter(active_flag=True) \
            .filter(available_on_ecom=True) \
            .filter(product__active_flag=True) \
            .filter(product__category__active_flag=True) \
            .filter(product=variant.product) \
            .annotate(selected_variant=Case(
                When(id=variant.id, then=True),
                default=Value(False),
                output_field=BooleanField(),
            ))

    if order_by is not None:
        pv = pv.order_by(order_by)

    return pv



from django.contrib.sessions.models import Session
from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone

from saleboxdjango.models import BasketWishlist


class Command(BaseCommand):
    def handle(self, *args, **options):
        # delete baskets from expired sessions
        expired_keys = Session \
                        .objects \
                        .filter(expire_date__lte=timezone.now()) \
                        .values_list('session_key', flat=True)

        BasketWishlist \
            .objects \
            .filter(user__isnull=True) \
            .filter(session__in=list(expired_keys)) \
            .delete()

        # delete baskets from deleted sessions
        remaining_keys = Session \
                            .objects \
                            .values_list('session_key', flat=True)

        BasketWishlist \
            .objects \
            .filter(user__isnull=True) \
            .exclude(session__in=list(remaining_keys)) \
            .delete()



from django import forms


class SaleboxBaseForm(forms.Form):
    redirect = forms.CharField(required=False)
    results = forms.CharField(required=False)
    state = forms.CharField(required=False)


# address
class SaleboxAddressAddForm(SaleboxBaseForm):
    default = forms.BooleanField(required=False)
    address_group = forms.CharField(required=False)
    full_name = forms.CharField(max_length=150)
    address_1 = forms.CharField(max_length=150)
    address_2 = forms.CharField(max_length=150, required=False)
    address_3 = forms.CharField(max_length=150, required=False)
    address_4 = forms.CharField(max_length=150, required=False)
    address_5 = forms.CharField(max_length=150, required=False)
    country_state = forms.IntegerField(required=False)
    country = forms.IntegerField(required=False)
    postcode = forms.CharField(max_length=12, required=False)
    phone_1 = forms.CharField(max_length=20, required=False)
    phone_2 = forms.CharField(max_length=20, required=False)
    email = forms.EmailField(required=False)
    string_1 = forms.CharField(max_length=255, required=False)
    string_2 = forms.CharField(max_length=255, required=False)
    tax_id = forms.CharField(max_length=30, required=False)


class SaleboxAddressIDForm(SaleboxBaseForm):
    id = forms.IntegerField()


# basket
class SaleboxBasketBasketForm(SaleboxBaseForm):
    variant_id = forms.IntegerField()
    quantity = forms.IntegerField()
    relative = forms.BooleanField(required=False)


class SaleboxBasketWishlistForm(SaleboxBaseForm):
    variant_id = forms.IntegerField()
    add = forms.BooleanField(required=False)


class SaleboxBasketMigrateForm(SaleboxBaseForm):
    variant_id = forms.IntegerField()
    to_basket = forms.BooleanField(required=False)


# rating
class SaleboxRatingAddForm(SaleboxBaseForm):
    variant_id = forms.IntegerField()
    rating = forms.IntegerField(min_value=0, max_value=100)
    review = forms.CharField(required=False)


class SaleboxRatingRemoveForm(SaleboxBaseForm):
    variant_id = forms.IntegerField()



from django.core.cache import cache

from saleboxdjango.models import Product, ProductCategory, ProductVariant


class SaleboxCategory:
    def __init__(self):
        self.populated_categories_only = True
        self.product_attribute_include = {}
        self.product_attribute_exclude = {}
        self.variant_attribute_include = {}
        self.variant_attribute_exclude = {}

        self.valid_ids = []

    def get_tree(self, cache_key=None, cache_timeout=86400, category_id=None):
        tree = None
        if cache_key is not None:
            tree = cache.get(cache_key)

        # build tree
        if tree is None:
            tree = self._get_tree(self._get_root_categories())
            if cache_key is not None:
                cache.set(cache_key, tree, cache_timeout)

        # create output
        current = self._find_node(tree, category_id) if category_id else None
        output = {
            'tree': tree,
            'current': current,
            'ancestors': self._find_ancestors(tree, [], current['parent'] if current else None)
        }

        return output

    def _find_ancestors(self, tree, ancestors, id):
        if id is not None:
            curr = self._find_node(tree, id)
            del curr['children']
            if curr is not None:
                ancestors.append(curr)
                if curr['parent'] is not None:
                    ancestors = self._find_ancestors(tree, ancestors, curr['parent'])

        return ancestors

    def _find_node(self, tree, id):
        for c in tree:
            if id == c['id']:
                return c
            else:
                res = self._find_node(c['children'], id)
                if res is not None:
                    return res

    def _get_root_categories(self):
        tree = []
        categories = ProductCategory \
                        .objects \
                        .filter(active_flag=True) \
                        .order_by('name')
        for c in categories:
            if c.is_root_node():
                tree.append(c)

        return tree

    def _get_product_count(self, category):
        ids = category \
                .get_descendants(include_self=True) \
                .values_list('id', flat=True)

        pv = ProductVariant \
                .objects \
                .filter(product__category__id__in=list(ids)) \
                .filter(product__active_flag=True) \
                .filter(active_flag=True) \
                .filter(available_on_ecom=True) \

        if len(self.product_attribute_include.keys()) > 0:
            pv = pv.filter(**self.product_attribute_include)

        if len(self.product_attribute_exclude.keys()) > 0:
            pv = pv.exclude(**self.product_attribute_exclude)

        if len(self.variant_attribute_include.keys()) > 0:
            pv = pv.filter(**self.variant_attribute_include)

        if len(self.variant_attribute_exclude.keys()) > 0:
            pv = pv.exclude(**self.variant_attribute_exclude)

        return pv \
                .order_by('product__id') \
                .distinct('product__id') \
                .count()

    def _get_tree(self, categories):
        output = []

        for c in categories:
            count = self._get_product_count(c)
            if self.populated_categories_only and count == 0:
                continue

            children = self._get_tree(
                c.get_children().filter(active_flag=True)
            )

            output.append({
                'id': c.id,
                'children': children,
                'i18n': c.i18n,
                'image': c.local_image,
                'name': c.name,
                'parent': c.parent.id if c.parent else None,
                'product_count': count,
                'short_name': c.short_name,
                'slug': c.slug,
                'slug_path': c.slug_path,
            })

        return output

    def set_product_attribute_include(self, attribute_number, value):
        key = 'product__attribute_%s' % attribute_number
        self.product_attribute_include[key] = value

    def set_product_attribute_include_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'product__attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.product_attribute_include[key] = field_value

    def set_product_attribute_exclude(self, attribute_number, value):
        key = 'product__attribute_%s' % attribute_number
        self.product_attribute_exclude[key] = value

    def set_product_attribute_exclude_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'product__attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.product_attribute_exclude[key] = field_value

    def set_variant_attribute_include(self, attribute_number, value):
        key = 'attribute_%s' % attribute_number
        self.variant_attribute_include[key] = value

    def set_variant_attribute_include_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.variant_attribute_include[key] = field_value

    def set_variant_attribute_exclude(self, attribute_number, value):
        key = 'attribute_%s' % attribute_number
        self.variant_attribute_exclude[key] = value

    def set_variant_attribute_exclude_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.variant_attribute_exclude[key] = field_value



from django.core.cache import cache
from saleboxdjango.models import Translation


def get_translation(lang, key, default='', translations=None):
    if translations is None:
        translations = get_translations()

    try:
        return translations[lang][key]
    except:
        if lang == 'en':
            try:
                return translations['en-US'][key]
            except:
                pass

        # fallback
        return default

def get_translations(rebuild=False):
    cache_key = 'salebox_translations'

    if rebuild:
        cache.delete(cache_key)

    # attempt to retrieve from cache
    o = cache.get(cache_key)
    if o is None:
        o = {}
        ts = Translation.objects.all()
        for t in ts:
            if t.language_code not in o:
                o[t.language_code] = {}
            o[t.language_code][t.key] = t.value
        cache.set(cache_key, o, 60 * 60 * 24 * 7)

    # return
    return o


from saleboxdjango.views.base import SaleboxBaseView
from saleboxdjango.lib.address import SaleboxAddress


class SaleboxAddressView(SaleboxBaseView):
    def init_class(self, request):
        self.sa = SaleboxAddress(request.user)



from random import randint
import datetime
import pytz
import time
import uuid

from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.contrib.postgres.fields import JSONField
from django.core.cache import cache
from django.db import models
from django.db.models import Avg
from django.utils.timezone import now
from django.utils.translation import get_language

from mptt.models import MPTTModel, TreeForeignKey
from saleboxdjango.lib.common import get_rating_display, json_to_datetime_local, json_to_datetime_utc


CHECKOUT_STATUS_CHOICES = (
    (10, 'New: Pending send to gateway'),
    (20, 'Pending: Awaiting gateway response'),
    (30, 'Success: Pending POST to Salebox'),
    (31, 'Success: Successfully POSTed to Salebox'),
    (40, 'Rejected: Gateway rejected payment'),
    (50, 'Timeout: Gateway did not respond in an acceptable time period')
)

class Attribute(models.Model):
    code = models.CharField(max_length=20)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.code

    class Meta:
        ordering = ['code']
        verbose_name = 'Product Attribute'

    def delete(self):
        pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class AttributeItem(models.Model):
    attribute = models.ForeignKey(Attribute, blank=True, null=True, on_delete=models.CASCADE)
    value = models.CharField(max_length=100)
    slug = models.CharField(max_length=100, blank=True, db_index=True)
    created = models.DateTimeField(auto_now_add=True)
    delete_dt = models.DateTimeField(blank=True, null=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.value

    class Meta:
        ordering = ['value']
        verbose_name = 'Product Attribute Item'

    def delete(self):
        pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class BasketWishlist(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.CASCADE)
    session = models.CharField(max_length=32, blank=True, null=True)
    basket_flag = models.BooleanField(default=True)
    variant = models.ForeignKey('ProductVariant', on_delete=models.CASCADE)
    quantity = models.IntegerField(default=1)
    weight = models.IntegerField(default=0)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

class CallbackStore(models.Model):
    ip_address = models.GenericIPAddressField()
    method = models.CharField(max_length=7)
    post = JSONField(blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)

class CheckoutStore(models.Model):
    uuid = models.UUIDField(db_index=True)
    visible_id = models.CharField(max_length=14, unique=True)
    user = models.IntegerField(blank=True, null=True)
    gateway_code = models.CharField(max_length=12)
    status = models.IntegerField(choices=CHECKOUT_STATUS_CHOICES)
    data = JSONField()
    payment_received = models.DateTimeField(blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(auto_now=True)

    def check_for_timeout(self):
        if self.status < 30:
            age = (now() - self.created).total_seconds()
            sb = getattr(settings, 'SALEBOX', {})
            cutoff = sb.get('GATEWAY_TIMEOUT', 60 * 60 * 24 * 3)

            # do update if applicable
            if age > cutoff:
                self.status = 50
                self.save()
                return True

        # fallback
        return False

    def save(self, *args, **kwargs):
        if self.uuid is None:
            self.uuid = str(uuid.uuid4())

        if self.visible_id is None:
            self.visible_id = '%.2f' % time.time()
            self.visible_id = self.visible_id.replace('.', '')
            chars = 'BCDFGHJKLMNPQRSTWXY'
            while len(self.visible_id) < 14:
                self.visible_id = '%s%s' % (
                    self.visible_id,
                    chars[randint(1, len(chars)) - 1]
                )

        super().save(*args, **kwargs)

        # update inventory counts
        exclude_ids = []
        variants = {}

        # fetch all pending checkouts
        checkouts = CheckoutStore.objects.filter(status__lt=30)
        for c in checkouts:
            for b in c.data['basket']['items']:
                pvid = b['variant']['id']
                if pvid not in variants:
                    variants[pvid] = 0
                variants[pvid] += b['qty']

        # update stock levels for all products
        for v in variants:
            pv = ProductVariant.objects.filter(id=v).first()
            if pv:
                exclude_ids.append(pv.id)
                pv.set_stock_checked_out(variants[v])

        # "reset" all the other checked out stock levels
        pvs = ProductVariant \
                .objects \
                .exclude(id__in=exclude_ids) \
                .filter(stock_checked_out__gt=0)
        for pv in pvs:
            pv.set_stock_checked_out(0)

    def set_status(self, status):
        if status == 30:
            if settings.SALEBOX['CHECKOUT']['TRANSACTION_DATE'] == 'payment':
                self.payment_received = datetime.datetime.now(tz=pytz.utc)
        self.status = status
        self.save()

class CheckoutStoreUpdate(models.Model):
    store = models.ForeignKey(CheckoutStore, on_delete=models.CASCADE)
    status = models.IntegerField(choices=CHECKOUT_STATUS_CHOICES)
    data = JSONField()
    created = models.DateTimeField(auto_now_add=True)

class Country(models.Model):
    code = models.CharField(max_length=2, blank=True, null=True)
    default = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.code

    class Meta:
        ordering = ['code']
        verbose_name_plural = 'Countries'

class CountryState(models.Model):
    country = models.ForeignKey(Country, blank=True, null=True, on_delete=models.CASCADE)
    code = models.CharField(max_length=5, blank=True, null=True)
    full_code = models.CharField(max_length=8, blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.full_code

    class Meta:
        ordering = ['full_code']
        verbose_name_plural = 'Country States'

class DiscountGroup(models.Model):
    GROUP_TYPE_CHOICES = (
        ('S', 'Seasonal'),
        ('M', 'Manual')
    )
    FIELD_CHOICES = (
        (0, 'Inactive'),
        (1, 'Active, not required'),
        (2, 'Active, required'),
    )
    ROUNDING_CHOICES = (
        ('none', 'No rounding'),
        ('up_major_1', 'Round UP to nearest major unit'),
        ('up_major_5', 'Round UP to nearest major unit multiple of 5'),
        ('up_major_10', 'Round UP to nearest major unit multiple of 10'),
    )
    name = models.CharField(max_length=25)
    group_type = models.CharField(max_length=1, default='M', choices=GROUP_TYPE_CHOICES)
    operational_flag = models.BooleanField(default=False)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = 'Discount Group'

    def delete(self):
        pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class DiscountRuleset(models.Model):
    TYPE_CHOICES = (
        ('flat_percent', 'Flat Percentage'),
    )
    group = models.ForeignKey(DiscountGroup, blank=True, null=True, on_delete=models.CASCADE)
    name = models.CharField(max_length=25)
    discount_type = models.CharField(default='flat_percent', max_length=12, choices=TYPE_CHOICES)
    value = models.IntegerField(null=True, blank=True)
    start_dt = models.DateTimeField(null=True, blank=True)
    end_dt = models.DateTimeField(null=True, blank=True)
    product_variant = models.ManyToManyField('ProductVariant', blank=True)
    operational_flag = models.BooleanField(default=True)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = 'Discount Ruleset'

    def delete(self):
        pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class Event(models.Model):
    event = models.CharField(max_length=50)
    transaction_guid = models.CharField(max_length=20, blank=True, null=True)
    salebox_member_id = models.UUIDField(default=uuid.uuid4, blank=True, null=True)
    processed_flag = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ['-created']

    def save(self, *args, **kwargs):
        # do not add duplicates
        if self.id is None:
            duplicate = Event \
                            .objects \
                            .filter(event=self.event) \
                            .filter(transaction_guid=self.transaction_guid) \
                            .filter(salebox_member_id=self.salebox_member_id) \
                            .filter(processed_flag=False) \
                            .count() > 0
            if duplicate:
                return
        super().save(*args, **kwargs)

class LastUpdate(models.Model):
    code = models.CharField(max_length=36)
    value = models.FloatField(default=0.0)

    def __str__(self):
        return self.code

    class Meta:
        ordering = ['code']
        verbose_name = 'Last Update'

class MemberGroup(models.Model):
    name = models.CharField(max_length=50)
    flat_discount_percentage = models.FloatField(default=0)
    can_be_parent = models.BooleanField(default=True)
    default_group = models.BooleanField(default=False)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = 'Member Group'

    def delete(self):
        pass

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class Member(models.Model):
    TITLE_CHOICES = (
        (1, 'Mr'),
        (2, 'Mrs'),
        (3, 'Miss'),
    )
    GENDER_CHOICES = (
        ('M', 'Male'),
        ('F', 'Female'),
        ('U', 'Unspecified'),
    )
    STATUS_CHOICES = (
        ('A', 'Active'),
        ('R', 'Resigned'),
        ('S', 'Suspended'),
        ('T', 'Terminated'),
    )
    # id = models.IntegerField(primary_key=True)
    group = models.ForeignKey(MemberGroup, null=True, blank=True, on_delete=models.CASCADE)
    parent = models.ForeignKey('Member', null=True, blank=True, on_delete=models.CASCADE)
    guid = models.CharField(max_length=25, db_index=True)
    salebox_member_id = models.UUIDField(
        blank=True,
        db_index=True,
        editable=True,
        null=True
    )
    gender = models.CharField(max_length=1, blank=True, null=True, choices=GENDER_CHOICES)
    title = models.IntegerField(blank=True, null=True, choices=TITLE_CHOICES)
    name_first = models.CharField(max_length=20, blank=True, null=True)
    name_last = models.CharField(max_length=30, blank=True, null=True)
    date_of_birth = models.DateField(blank=True, null=True)
    address_1 = models.CharField(max_length=255, blank=True, null=True)
    address_2 = models.CharField(max_length=255, blank=True, null=True)
    address_3 = models.CharField(max_length=255, blank=True, null=True)
    address_4 = models.CharField(max_length=255, blank=True, null=True)
    address_5 = models.CharField(max_length=255, blank=True, null=True)
    postcode = models.CharField(max_length=12, blank=True, null=True)
    country = models.ForeignKey(Country, blank=True, null=True, on_delete=models.CASCADE)
    country_state = models.ForeignKey(CountryState, blank=True, null=True, on_delete=models.CASCADE)
    phone_1 = models.CharField(max_length=20, blank=True, null=True)
    phone_2 = models.CharField(max_length=20, blank=True, null=True)
    email = models.EmailField(blank=True, null=True)
    id_card = models.CharField(max_length=20, blank=True, null=True)
    status = models.CharField(max_length=1, default='A', choices=STATUS_CHOICES)
    active_flag = models.BooleanField(default=True)
    join_date = models.DateField(blank=True, null=True)
    string_1 = models.CharField(max_length=255, blank=True, null=True)
    string_2 = models.CharField(max_length=255, blank=True, null=True)
    string_3 = models.CharField(max_length=255, blank=True, null=True)
    string_4 = models.CharField(max_length=255, blank=True, null=True)
    string_5 = models.CharField(max_length=255, blank=True, null=True)
    string_6 = models.CharField(max_length=255, blank=True, null=True)
    string_7 = models.CharField(max_length=255, blank=True, null=True)
    string_8 = models.CharField(max_length=255, blank=True, null=True)
    string_9 = models.CharField(max_length=255, blank=True, null=True)
    string_10 = models.CharField(max_length=255, blank=True, null=True)
    string_11 = models.CharField(max_length=255, blank=True, null=True)
    string_12 = models.CharField(max_length=255, blank=True, null=True)
    string_13 = models.CharField(max_length=255, blank=True, null=True)
    string_14 = models.CharField(max_length=255, blank=True, null=True)
    boolean_1 = models.BooleanField(default=False)
    boolean_2 = models.BooleanField(default=False)
    boolean_3 = models.BooleanField(default=False)
    boolean_4 = models.BooleanField(default=False)
    boolean_5 = models.BooleanField(default=False)
    boolean_6 = models.BooleanField(default=False)
    group_when_created = models.ForeignKey(MemberGroup, blank=True, null=True, related_name='group_when_created', on_delete=models.CASCADE)

    salebox_transactionhistory_request_dt = models.DateTimeField(default=datetime.datetime.now, blank=True, null=True)
    salebox_transactionhistory_count = models.IntegerField(default=0)
    salebox_transactionhistory_data = JSONField(blank=True, null=True)

    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.guid

    class Meta:
        ordering = ['guid']
        verbose_name = 'Member'

    def delete(self):
        pass

    def transactionhistory_request_sync(self):
        self.salebox_transactionhistory_request_dt = datetime.datetime.now()
        self.save()

    def transactionhistory_update_data(self, transaction_list):
        for t in transaction_list:
            del t['member']
        self.salebox_transactionhistory_data = transaction_list
        self.salebox_transactionhistory_count = len(transaction_list)
        self.salebox_transactionhistory_request_dt = None
        self.save()

    def transactionhistory_get_data(self):
        transactions = self.salebox_transactionhistory_data
        for t in transactions:
            t['dt'] = json_to_datetime_local(t['dt'])
            t['basket_size'] = 0
            for b in t['basket']:
                try:
                    pv = ProductVariant \
                            .objects \
                            .filter(id=b['product_variant']['id']) \
                            .select_related('product', 'product__category') \
                            .first()
                    b['product_variant'] = pv
                    b['product'] = pv.product
                    b['category'] = pv.product.category
                except:
                    pass
                t['basket_size'] += b['quantity']

        return transactions

    def transactionhistory_get_single_transaction(self, pos_guid):
        pos_guid = pos_guid.lower()
        for t in self.salebox_transactionhistory_data:
            if t['pos_guid'].lower() == pos_guid:
                return t

        return None

class ProductCategory(MPTTModel):
    short_name = models.CharField(max_length=30)
    name = models.CharField(max_length=100)
    parent = TreeForeignKey('self', null=True, blank=True, related_name='children', db_index=True, on_delete=models.CASCADE)
    image = models.CharField(max_length=70, blank=True, null=True)
    local_image = models.CharField(max_length=25, blank=True, null=True)
    seasonal_flag = models.BooleanField(default=False)
    slug = models.CharField(max_length=100, blank=True, null=True, db_index=True)
    slug_path = models.CharField(max_length=255, blank=True, null=True, db_index=True)
    active_flag = models.BooleanField(default=True)
    i18n = JSONField(default=dict)
    boolean_1 = models.BooleanField(default=False)
    boolean_2 = models.BooleanField(default=False)
    boolean_3 = models.BooleanField(default=False)
    boolean_4 = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = 'Product Category'
        verbose_name_plural = 'Product Categories'

    def delete(self):
        pass

    def get_trans_for_lang(self, lang, field, fallback_to_primary=True, non_primary_fallback=None):
        if not hasattr(self, field):
            return '[BAD FIELD NAME]'

        i18n = self.i18n.get(lang, {})
        if field in i18n:
            return i18n[field]

        # fallback
        if fallback_to_primary:
            return getattr(self, field)
        else:
            return non_primary_fallback

    def get_trans(self, field, fallback_to_primary=True, non_primary_fallback=None):
        return self.get_trans_for_lang(get_language(), field, fallback_to_primary, non_primary_fallback)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        root = self.get_root()
        nodes = root.get_descendants(include_self=True)
        for node in nodes:
            slugs = node.get_ancestors(include_self=True).values_list('slug', flat=True)
            slugs = ['' if s is None else s for s in slugs]
            slug_path = '/'.join(slugs)
            if node.slug_path != slug_path:
                ProductCategory.objects.filter(id=node.id).update(slug_path=slug_path)
        cache.clear()

class Product(models.Model):
    SOLD_BY_CHOICES = (
        ('item', 'item'),
        # ('volume', 'volume'),
        ('weight', 'weight'),
    )
    category = models.ForeignKey(ProductCategory, blank=True, null=True,on_delete=models.CASCADE)
    name = models.CharField(max_length=150)
    attribute_1 = models.ManyToManyField(AttributeItem, related_name='product_attr_1', blank=True)
    attribute_2 = models.ManyToManyField(AttributeItem, related_name='product_attr_2', blank=True)
    attribute_3 = models.ManyToManyField(AttributeItem, related_name='product_attr_3', blank=True)
    attribute_4 = models.ManyToManyField(AttributeItem, related_name='product_attr_4', blank=True)
    attribute_5 = models.ManyToManyField(AttributeItem, related_name='product_attr_5', blank=True)
    attribute_6 = models.ManyToManyField(AttributeItem, related_name='product_attr_6', blank=True)
    attribute_7 = models.ManyToManyField(AttributeItem, related_name='product_attr_7', blank=True)
    attribute_8 = models.ManyToManyField(AttributeItem, related_name='product_attr_8', blank=True)
    attribute_9 = models.ManyToManyField(AttributeItem, related_name='product_attr_9', blank=True)
    attribute_10 = models.ManyToManyField(AttributeItem, related_name='product_attr_10', blank=True)
    string_1 = models.CharField(max_length=150, blank=True, null=True)
    string_2 = models.CharField(max_length=150, blank=True, null=True)
    string_3 = models.CharField(max_length=150, blank=True, null=True)
    string_4 = models.CharField(max_length=150, blank=True, null=True)
    boolean_1 = models.BooleanField(default=False)
    boolean_2 = models.BooleanField(default=False)
    boolean_3 = models.BooleanField(default=False)
    boolean_4 = models.BooleanField(default=False)
    sold_by = models.CharField(max_length=6, choices=SOLD_BY_CHOICES, default='item')
    vat_applicable = models.BooleanField(default=True)
    image = models.CharField(max_length=70, blank=True, null=True)
    local_image = models.CharField(max_length=25, blank=True, null=True)
    inventory_flag = models.BooleanField(default=True)
    # season = models.ForeignKey(OrganizationSeason, null=True, blank=True)
    slug = models.CharField(max_length=100, blank=True, null=True, db_index=True)
    bestseller_rank = models.IntegerField(default=0)
    active_flag = models.BooleanField(default=True)
    i18n = JSONField(default=dict)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    # local values
    rating_score = models.IntegerField(default=0)
    rating_vote_count = models.IntegerField(default=0)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = 'Product'

    def delete(self):
        pass

    def get_trans_for_lang(self, lang, field, fallback_to_primary=True, non_primary_fallback=None):
        if not hasattr(self, field):
            return '[BAD FIELD NAME]'

        i18n = self.i18n.get(lang, {})
        if field in i18n:
            return i18n[field]

        # fallback
        if fallback_to_primary:
            return getattr(self, field)
        else:
            return non_primary_fallback

    def get_trans(self, field, fallback_to_primary=True, non_primary_fallback=None):
        return self.get_trans_for_lang(get_language(), field, fallback_to_primary, non_primary_fallback)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

    def update_rating(self):
        variant_ids = ProductVariant \
                        .objects \
                        .filter(product=self) \
                        .values_list('id', flat=True)

        self.rating_score = 0
        self.rating_vote_count = ProductVariantRating \
                                    .objects \
                                    .filter(variant__in=variant_ids) \
                                    .count()

        if self.rating_vote_count > 0:
            self.rating_score = ProductVariantRating \
                                    .objects \
                                    .filter(variant__in=variant_ids) \
                                    .aggregate(rating=Avg('rating'))['rating']

        self.save()

class ProductVariant(models.Model):
    SHELF_EXPIRY_CHOICES = (
        ('none', 'No shelf expiry date'),
        ('manual', 'Enter shelf expiry date manually'),
        ('delivery', 'Calculate from delivery date'),
        ('manufacture', 'Calculate from manufacture date'),
    )
    SIZE_UOM_CHOICES = (
        ('', 'n/a'),
        ('g', 'g'),
        ('kg', 'kg'),
        ('ml', 'ml'),
    )
    product = models.ForeignKey(Product, blank=True, null=True, on_delete=models.CASCADE)
    name = models.CharField(max_length=150, blank=True, default='')
    bo_name = models.CharField(max_length=200, blank=True, default='')
    plu = models.CharField(max_length=25, blank=True, default='')
    sku = models.CharField(max_length=25, blank=True, default='')
    color = models.CharField(max_length=50, blank=True, default='')
    size = models.CharField(max_length=20, blank=True, default='')
    size_order = models.FloatField(default=0)
    size_uom = models.CharField(max_length=2, choices=SIZE_UOM_CHOICES, blank=True, default='')
    price = models.IntegerField(null=True)
    sale_price = models.IntegerField(default=0)
    sale_percent = models.IntegerField(default=0)
    barcode = models.CharField(max_length=50, blank=True, default='')
    available_to_order = models.BooleanField(default=True)
    available_on_pos = models.BooleanField(default=True)
    available_on_ecom = models.BooleanField(default=True)
    shelf_expiry_type = models.CharField(max_length=12, default='manual')
    shelf_life_days = models.IntegerField(blank=True, null=True)
    slug = models.CharField(max_length=150, blank=True, null=True, db_index=True)
    image = models.CharField(max_length=70, blank=True, null=True)
    local_image = models.CharField(max_length=20, blank=True, null=True)
    unique_string = models.CharField(max_length=255, blank=True)
    shipping_weight = models.IntegerField(blank=True, null=True)
    loyalty_points = models.FloatField(blank=True, null=True)
    member_discount_applicable = models.BooleanField(default=True)
    string_1 = models.CharField(max_length=150, blank=True, null=True)
    string_2 = models.CharField(max_length=150, blank=True, null=True)
    string_3 = models.CharField(max_length=150, blank=True, null=True)
    string_4 = models.CharField(max_length=150, blank=True, null=True)
    warehouse_location = models.CharField(max_length=50, blank=True, null=True)
    int_1 = models.IntegerField(blank=True, null=True)
    int_2 = models.IntegerField(blank=True, null=True)
    int_3 = models.IntegerField(blank=True, null=True)
    int_4 = models.IntegerField(blank=True, null=True)
    date_1 = models.DateField(blank=True, null=True)
    date_2 = models.DateField(blank=True, null=True)
    boolean_1 = models.BooleanField(default=False)
    boolean_2 = models.BooleanField(default=False)
    boolean_3 = models.BooleanField(default=False)
    boolean_4 = models.BooleanField(default=False)
    attribute_1 = models.ManyToManyField(AttributeItem, related_name='variant_attr_1', blank=True)
    attribute_2 = models.ManyToManyField(AttributeItem, related_name='variant_attr_2', blank=True)
    attribute_3 = models.ManyToManyField(AttributeItem, related_name='variant_attr_3', blank=True)
    attribute_4 = models.ManyToManyField(AttributeItem, related_name='variant_attr_4', blank=True)
    attribute_5 = models.ManyToManyField(AttributeItem, related_name='variant_attr_5', blank=True)
    attribute_6 = models.ManyToManyField(AttributeItem, related_name='variant_attr_6', blank=True)
    attribute_7 = models.ManyToManyField(AttributeItem, related_name='variant_attr_7', blank=True)
    attribute_8 = models.ManyToManyField(AttributeItem, related_name='variant_attr_8', blank=True)
    attribute_9 = models.ManyToManyField(AttributeItem, related_name='variant_attr_9', blank=True)
    attribute_10 = models.ManyToManyField(AttributeItem, related_name='variant_attr_10', blank=True)
    active_flag = models.BooleanField(default=True)
    ecommerce_description = models.TextField(blank=True, null=True)
    ecommerce_low_stock_threshold = models.IntegerField(default=10)
    bestseller_rank = models.IntegerField(default=0)
    default_image = models.CharField(max_length=35, blank=True, null=True)
    i18n = JSONField(default=dict)
    preorder_flag = models.BooleanField(default=False)
    stock_count = models.IntegerField(default=0)
    stock_checked_out = models.IntegerField(default=0)
    stock_total = models.IntegerField(default=0)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    # local values
    rating_score = models.IntegerField(default=0)
    rating_vote_count = models.IntegerField(default=0)
    name_sorted = models.IntegerField(default=0, db_index=True)

    def __str__(self):
        return self.name or ''

    class Meta:
        ordering = ['id']
        indexes = (
            models.Index(fields=('bestseller_rank', 'name_sorted')),
            models.Index(fields=('-bestseller_rank', 'name_sorted')),
            models.Index(fields=('sale_price', 'name_sorted')),
            models.Index(fields=('-sale_price', 'name_sorted')),
            models.Index(fields=('rating_score', 'name_sorted')),
            models.Index(fields=('-rating_score', 'name_sorted')),
        )
        verbose_name = 'Product Variant'

        # CREATE INDEX <name_of_index> ON saleboxdjango_productvariant (col_a ASC, col_b DESC)

    def delete(self):
        pass

    def get_trans_for_lang(self, lang, field, fallback_to_primary=True, non_primary_fallback=None):
        if not hasattr(self, field):
            return '[BAD FIELD NAME]'

        i18n = self.i18n.get(lang, {})
        if field in i18n:
            return i18n[field]

        # fallback
        if fallback_to_primary:
            return getattr(self, field)
        else:
            return non_primary_fallback

    def get_trans(self, field, fallback_to_primary=True, non_primary_fallback=None):
        return self.get_trans_for_lang(get_language(), field, fallback_to_primary, non_primary_fallback)

    def rating_display(self):
        return get_rating_display(self.rating_score, self.rating_vote_count)

    def save(self, *args, **kwargs):
        self.default_image = None

        # attempt to use first variant image
        pvi = ProductVariantImage \
                .objects \
                .filter(variant=self) \
                .filter(local_img__isnull=False) \
                .filter(active_flag=True) \
                .order_by('order') \
                .first()
        if pvi is not None:
            self.default_image = 'pvi/%s' % pvi.local_img

        # fallback to POS image if that exists
        if self.default_image is None and self.local_image is not None:
            self.default_image = 'pospv/%s' % self.local_image

        # set stock_total
        self.stock_total = max((self.stock_count - self.stock_checked_out), 0)

        # remove basket/wishlist item if varaint now unavailable
        if not self.available_on_ecom or not self.active_flag:
            BasketWishlist.objects.filter(variant=self).delete()

        # save
        super(ProductVariant, self).save(*args, **kwargs)
        cache.clear()

    def set_stock_checked_out(self, qty=0):
        if self.id:
            ProductVariant \
                .objects \
                .filter(id=self.id) \
                .update(
                    stock_checked_out=qty,
                    stock_total=max((self.stock_count - qty), 0)
                )

    def update_rating(self):
        self.rating_score = 0
        self.rating_vote_count = \
            ProductVariantRating \
                .objects \
                .filter(variant=self) \
                .count()

        if self.rating_vote_count > 0:
            self.rating_score = \
                ProductVariantRating \
                    .objects \
                    .filter(variant=self) \
                    .aggregate(rating=Avg('rating'))['rating']

        self.save()
        self.product.update_rating()

class ProductVariantImage(models.Model):
    variant = models.ForeignKey(ProductVariant, on_delete=models.CASCADE)
    img = models.CharField(max_length=100, default='')
    local_img = models.CharField(max_length=25, blank=True, null=True)
    img_height = models.IntegerField(default=0)
    img_width = models.IntegerField(default=0)
    title = models.CharField(max_length=150, blank=True, null=True)
    description = models.CharField(max_length=255, blank=True, null=True)
    order = models.IntegerField(blank=True, null=True)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ('order',)

    def save(self, *args, **kwargs):
        super(ProductVariantImage, self).save(*args, **kwargs)
        self.variant.save()

class ProductVariantRating(models.Model):
    variant = models.ForeignKey('ProductVariant', on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    rating = models.IntegerField(default=50)
    review = models.TextField(blank=True, null=True)
    review_qc_flag = models.BooleanField(default=False)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = 'Product Variant Rating'

    def delete(self, *args, **kwargs):
        variant = self.variant
        super().delete(*args, **kwargs)
        variant.update_rating()

    def save(self, *args, **kwargs):
        if self.review is not None:
            self.review = self.review.strip()
            if len(self.review) == 0:
                self.review = None
        super().save(*args, **kwargs)
        self.variant.update_rating()

    def rating_10(self):
        try:
            return round(self.rating / 10)
        except:
            return None

    def rating_5(self):
        try:
            return round(self.rating / 20)
        except:
            return None

class SaleboxUser(AbstractUser):
    salebox_member_id = models.UUIDField(blank=True, db_index=True, editable=True, null=True)
    salebox_member_sync = JSONField(blank=True, null=True)
    salebox_last_update = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def create_salebox_member_id(self):
        if self.salebox_member_id is None:
            self.salebox_member_id = str(uuid.uuid4())
            self.save()

    def get_member(self):
        if self.salebox_member_id is None:
            return None

        return Member \
                .objects \
                .filter(salebox_member_id=self.salebox_member_id) \
                .first()

    def set_salebox_member_sync(self, key, value):
        if self.salebox_member_sync is None:
            self.salebox_member_sync = {}
        self.salebox_member_sync[key] = value
        self.save()

class Translation(models.Model):
    language_code = models.CharField(max_length=7)
    key = models.CharField(max_length=255)
    value = models.TextField(blank=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['language_code', 'key']
        verbose_name = 'Translation'

    def delete(self):
        super().delete(*args, **kwargs)
        cache.clear()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.clear()

class UserAddress(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    default = models.BooleanField(default=False)
    address_group = models.CharField(max_length=10, default='default')
    full_name = models.CharField(max_length=150)
    address_1 = models.CharField(max_length=150)
    address_2 = models.CharField(max_length=150, blank=True, null=True)
    address_3 = models.CharField(max_length=150, blank=True, null=True)
    address_4 = models.CharField(max_length=150, blank=True, null=True)
    address_5 = models.CharField(max_length=150, blank=True, null=True)
    country_state = models.ForeignKey(CountryState, blank=True, null=True, on_delete=models.CASCADE)
    country = models.ForeignKey(Country, blank=True, null=True, on_delete=models.CASCADE)
    postcode = models.CharField(max_length=12, blank=True, null=True)
    phone_1 = models.CharField(max_length=20, blank=True, null=True)
    phone_2 = models.CharField(max_length=20, blank=True, null=True)
    email = models.EmailField(blank=True, null=True)
    string_1 = models.CharField(max_length=255, blank=True, null=True)
    string_2 = models.CharField(max_length=255, blank=True, null=True)
    tax_id = models.CharField(max_length=36, blank=True, null=True)
    active_flag = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return '%s, %s...' % (self.full_name, self.address_1)

    class Meta:
        ordering = ['-default', 'full_name', 'address_1']
        verbose_name = 'User Address'

    def delete(self, *args, **kwargs):
        user = self.user
        address_group = self.address_group
        super().delete(*args, **kwargs)
        self.ensure_one_default_exists(user, address_group)

    def save(self, *args, **kwargs):
        if self.tax_id is not None:
            self.tax_id = self.tax_id.strip()
            if len(self.tax_id) == 0:
                self.tax_id = None
        super().save(*args, **kwargs)
        self.ensure_one_default_exists(self.user, self.address_group)

    def ensure_one_default_exists(self, user, address_group):
        addresses = UserAddress \
                        .objects \
                        .filter(user=user) \
                        .filter(address_group=address_group) \
                        .order_by('-last_update')

        if len(addresses) > 0:
            default_count = 0

            # ensure in inactive addresses are marked as default
            for a in addresses.all():
                if not a.active_flag and a.default:
                    UserAddress.objects.filter(id=a.id).update(default=False)

            # ensure we have a maximum of one default
            for a in addresses.all():
                if a.default:
                    default_count += 1
                    if default_count > 1:
                        UserAddress.objects.filter(id=a.id).update(default=False)

            # if we don't have a default, set one
            if default_count == 0:
                for a in addresses.all():
                    if a.active_flag:
                        UserAddress.objects.filter(id=a.id).update(default=True)
                        break



from saleboxdjango.views.basket.base import SaleboxBasketView
from saleboxdjango.forms import SaleboxBasketMigrateForm


class SaleboxBasketMigrateView(SaleboxBasketView):
    action = 'basket-migrate'
    form = SaleboxBasketMigrateForm

    def form_valid(self, request):
        try:
            self.sb.migrate_basket_wishlist(
                request,
                self.form.cleaned_data['variant_id'],
                self.form.cleaned_data['to_basket']
            )
            self.status = 'success'

            self.results = self.sb.get_data(
                request,
                self.results_csv
            )
        except:
            self.status = 'fail'



from pprint import pprint

from pyshipping.package import Package
from pyshipping.binpack_simple import binpack


class BinPack:
    def __init__(self):
        self.bins = []
        self.bin_lookup = {}
        self.packages = []
        self.oversize = []
        self.min_bin_size = -1
        self.options = []

    def add_bin(self, id, price, height, width, depth):
        # add bin
        pkg = Package((height, width, depth))
        self.bins.append({
            'id': id,
            'price': price,
            'pkg': pkg,
        })

        # sort bins, cheapest first
        self.bins = sorted(self.bins, key=lambda x: (x['price']))

        # update bin lookup
        for i, b in enumerate(self.bins):
            self.bin_lookup[b['id']] = i

    def add_package(self, meta, height, width, depth):
        pkg = Package((height, width, depth))

        # set smallest container size
        oversize = True
        for i, b in enumerate(self.bins):
            good, bad = binpack([pkg], b['pkg'])
            if len(good) == 1:
                oversize = False
                if i > self.min_bin_size:
                    self.min_bin_size = i
                break

        # add to appropriate list
        if oversize:
            self.oversize.append({
                'meta': meta,
                'pkg': pkg
            })
        else:
            self.packages.append({
                'meta': meta,
                'min_bin_size': i,
                'pkg': pkg,
            })

        # sort packages, largest volume first
        self.packages = sorted(
            self.packages,
            key=lambda x: x['pkg'].volume,
            reverse=True
        )

    def calculate(self, bin_sizes):
        containers = []

        for p in self.packages:
            containers.append({
                'contents': {
                    'meta': [],
                    'packages': [],
                },
                'id': self.bins[bin_sizes[-1]]['id'],
            })

        # loop through packages
        for p in self.packages:
            for c in containers:
                good, bad = binpack(
                    c['contents']['packages'] + [p['pkg']],
                    self.bins[self.bin_lookup[c['id']]]['pkg']
                )
                if len(good) == 1:
                    c['contents']['packages'] += [p['pkg']]
                    c['contents']['meta'] += [p['meta']]
                    break

        # remove empty bins
        for i in reversed(range(len(containers))):
            if len(containers[i]['contents']['packages']) == 0:
                containers.pop()

        # attempt to reduce the size of the bins
        for c in containers:
            for b in self.bins:
                good, bad = binpack(c['contents']['packages'], b['pkg'])
                if len(good) == 1 and len(bad) == 0:
                    c['id'] = b['id']
                    break

        return containers

    def go(self):
        tmp = []
        results = []

        # calculate bin size combinations
        bin_sizes = list(range(self.min_bin_size, len(self.bins)))
        while len(bin_sizes) > 0:
            tmp.append(self.calculate(bin_sizes))
            bin_sizes.pop()

        # calculate prices
        for t in tmp:
            total_price = 0
            for c in t:
                c['price'] = self.bins[self.bin_lookup[c['id']]]['price']
                total_price += c['price']

            results.append({
                'price': total_price,
                'number_of_packages': len(t),
                'packages': t
            })

        # remove duplicate options
        duplicates = []
        for r in results:
            s = '%s:%s' % (r['number_of_packages'], r['price'])
            if s not in duplicates:
                duplicates.append(s)
            else:
                r['duplicate'] = True
        results = [r for r in results if 'duplicate' not in r]

        # sort prices, cheapest first
        results = sorted(
            results,
            key=lambda x: (x['price'], x['number_of_packages']),
        )

        return {
            'results': results,
            'oversize': [o['meta'] for o in self.oversize]
        }

    def debug(self):
        pprint({
            'min_bin_size': self.min_bin_size,
            'bins': self.bins,
            'packages': self.packages,
            'oversize': self.oversize,
            'bin_lookup': self.bin_lookup
        })

"""
bp = BinPack()

bp.add_bin('S', 35, 140, 200, 60)
bp.add_bin('M', 65, 170, 250, 90)
bp.add_bin('L', 75, 200, 300, 110)

bp.add_package('A', 125, 180, 11)
bp.add_package('B', 125, 180, 12)
bp.add_package('C', 148, 209, 36)
bp.add_package('D', 208, 280, 12)

# bp.debug()
pprint(bp.go())
"""



import datetime
import pytz
import re

from django.conf import settings
from django.db import connection
from django.utils.timezone import localtime


def dictfetchall(sql):
    with connection.cursor() as cursor:
        cursor.execute(sql)
        columns = [col[0] for col in cursor.description]
        return [
            dict(zip(columns, row))
            for row in cursor.fetchall()
        ]

def fetchflatlist(sql):
    with connection.cursor() as cursor:
        cursor.execute(sql)
        return [row[0] for row in cursor.fetchall()]

def fetchsinglevalue(sql):
    return fetchflatlist(sql)[0]

def get_rating_display(score, vote_count):
    if vote_count > 0:
        return {
            '100': score,
            '5': round(score / 20),
            '10': round(score / 10)
        }
    else:
        return {'100': 0, '5': 0, '10': 0}

def json_to_datetime_local(value):
    return localtime(json_to_datetime_utc(value))

def json_to_datetime_utc(value):
    return datetime.datetime(*value, tzinfo=pytz.UTC)

def update_natural_sort(model, str_column, sort_column):
    # retrieve all rows
    rows = list(model.objects.values('id', str_column, sort_column))

    # update string to be sortable
    for row in rows:
        row[str_column] = row[str_column].lower()
        row[str_column] = row[str_column].strip()
        row[str_column] = re.sub(r'^the\s+', '', row[str_column])
        row[str_column] = re.sub(r'\d+', lambda x: '%08d' % (int(x.group(0)),), row[str_column])

    # sort the values
    rows = sorted(rows, key=lambda k: k[str_column])

    # update the database records that need it
    for i, row in enumerate(rows):
        if i != row[sort_column]:
            model.objects.filter(id=row['id']).update(**{sort_column: i})



from django.contrib import admin
from saleboxdjango.models import *


class CheckoutStoreUpdateInline(admin.TabularInline):
    model = CheckoutStoreUpdate
    readonly_fields = ('status', 'created', 'data')
    extra = 0

    def has_delete_permission(self, request, obj=None):
        return False

class DiscountRulesetInline(admin.TabularInline):
    model = DiscountRuleset
    exclude = ['product_variant']
    extra = 0


class AttributeAdmin(admin.ModelAdmin):
    list_display = ('code',)


class AttributeItemAdmin(admin.ModelAdmin):
    list_display = ('value', 'slug', 'attribute')
    list_filter = ('attribute',)


class BasketWishlistAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'session', 'variant', 'quantity', 'weight', 'basket_flag')
    list_filter = ('basket_flag',)
    search_fields = ('user__email',)


class CallbackStoreAdmin(admin.ModelAdmin):
    list_display = ('created', 'ip_address', 'method')
    search_fields = ('post',)


class CheckoutStoreAdmin(admin.ModelAdmin):
    list_display = ('created', 'status', 'uuid', 'visible_id', 'user', 'gateway_code')
    inlines = [CheckoutStoreUpdateInline]
    list_filter = ('status', 'gateway_code')
    readonly_fields = ('uuid', 'visible_id', 'user', 'gateway_code', 'status', 'data', 'payment_received')
    search_fields = ('visible_id', 'data')


class CountryAdmin(admin.ModelAdmin):
    list_display = ('code', 'default')


class CountryStateAdmin(admin.ModelAdmin):
    list_display = ('country', 'code', 'full_code')


class DiscountGroupAdmin(admin.ModelAdmin):
    inlines = [DiscountRulesetInline]
    list_display = ('name', 'group_type')
    list_filter = ('group_type',)


class EventAdmin(admin.ModelAdmin):
    list_display = ('event', 'salebox_member_id', 'transaction_guid', 'processed_flag', 'created')
    list_filter = ('event', 'processed_flag')
    readonly_fields = ('event', 'salebox_member_id', 'transaction_guid', 'created')
    search_fields = ('salebox_member_id', 'transaction_guid')


class LastUpdateAdmin(admin.ModelAdmin):
    list_display = ('code', 'value')


class MemberGroupAdmin(admin.ModelAdmin):
    list_display = ('name', 'flat_discount_percentage')


class MemberAdmin(admin.ModelAdmin):
    list_display = ('guid', 'email', 'name_first', 'name_last', 'group')
    list_filter = ('group',)
    readonly_fields = ('parent',)
    search_fields = ('guid', 'email')


class ProductCategoryAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug_path')
    search_fields = ('name', 'slug_path')


class ProductAdmin(admin.ModelAdmin):
    list_display = ('name', 'category', 'sold_by', 'slug')
    list_filter = ('sold_by', 'category')
    search_fields = ('name', 'slug', 'plu', 'sku')


class ProductVariantAdmin(admin.ModelAdmin):
    list_display = ('product', 'name', 'price', 'slug')
    list_filter = ('product__category',)


class ProductVariantRatingAdmin(admin.ModelAdmin):
    list_display = ('user', 'variant', 'rating', 'created')
    search_fields = ('user__email',)


class TranslationAdmin(admin.ModelAdmin):
    list_display = ('language_code', 'key', 'value')
    list_filter = ('language_code',)
    search_fields = ('key',)

class UserAddressAdmin(admin.ModelAdmin):
    list_display = (
        'address_group',
        'user',
        'full_name',
        'address_1',
        'address_2',
        'country_state',
        'country',
        'postcode'
    )
    list_filter = ('address_group',)
    search_fields = (
        'user__email',
        'full_name',
        'address_1',
        'address_2',
        'address_3',
        'postcode'
    )


admin.site.register(Attribute, AttributeAdmin)
admin.site.register(AttributeItem, AttributeItemAdmin)
admin.site.register(BasketWishlist, BasketWishlistAdmin)
admin.site.register(CallbackStore, CallbackStoreAdmin)
admin.site.register(CheckoutStore, CheckoutStoreAdmin)
admin.site.register(Country, CountryAdmin)
admin.site.register(CountryState, CountryStateAdmin)
admin.site.register(DiscountGroup, DiscountGroupAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(LastUpdate, LastUpdateAdmin)
admin.site.register(MemberGroup, MemberGroupAdmin)
admin.site.register(Member, MemberAdmin)
admin.site.register(ProductCategory, ProductCategoryAdmin)
admin.site.register(Product, ProductAdmin)
admin.site.register(ProductVariant, ProductVariantAdmin)
admin.site.register(ProductVariantRating, ProductVariantRatingAdmin)
admin.site.register(Translation, TranslationAdmin)
admin.site.register(UserAddress, UserAddressAdmin)



import time

from django.conf import settings
from django.forms.models import model_to_dict
from django.utils.translation import get_language

from saleboxdjango.models import CheckoutStore, UserAddress


class SaleboxCheckout:
    def __init__(self, request):
        self._init_sequence()
        self._init_session(request)
        self._write_session(request)

    def get_checkout_nav(self, curr_page_name=None):
        nav = {
            'order': [],
            'lookup': {},
            'previous': None,
            'next': None
        }

        for pos, s in enumerate(self.sequence['order']):
            accessible = curr_page_name is not None and self.sequence['lookup'][s]['accessible']
            current = s == curr_page_name

            # set previous / next item helpers
            if current:
                if pos > 0:
                    tmp = self.sequence['order'][pos - 1]
                    nav['previous'] = {
                        'accessible': self.sequence['lookup'][tmp]['accessible'],
                        'label': self.sequence['lookup'][tmp]['label'],
                        'page_name': tmp,
                        'path': self.sequence['lookup'][tmp]['path'],
                    }
                if pos < len(self.sequence['order']) - 1:
                    tmp = self.sequence['order'][pos + 1]
                    nav['next'] = {
                        'accessible': self.sequence['lookup'][tmp]['accessible'],
                        'label': self.sequence['lookup'][tmp]['label'],
                        'page_name': tmp,
                        'path': self.sequence['lookup'][tmp]['path'],
                    }

            # set lookup
            nav['lookup'][s] = {
                'accessible': accessible,
                'current': current,
            }

            # set order
            nav['order'].append({
                'accessible': accessible,
                'current': current,
                'label': self.sequence['lookup'][s]['label'],
                'page_name': s,
                'path': self.sequence['lookup'][s]['path'],
            })

        for pos, n in enumerate(nav['order']):
            try:
                nav['order'][pos]['previous_accessible'] = nav['order'][pos - 1]['accessible']
            except:
                nav['order'][pos]['previous_accessible'] = False
            try:
                nav['order'][pos]['next_accessible'] = nav['order'][pos + 1]['accessible']
            except:
                nav['order'][pos]['next_accessible'] = False

        return nav

    def get_raw_data(self):
        return self.data

    def get_last_accessible_page(self):
        for o in reversed(self.sequence['order']):
            if self.sequence['lookup'][o]['accessible']:
                return self.sequence['lookup'][o]['path']

        return None

    def get_current_page_path(self, path_name):
        return self.sequence['lookup'][path_name]['path']

    def get_next_page(self, page_name):
        next = self.sequence['order'].index(page_name) + 1
        if next < len(self.sequence['order']):
            return self.sequence['lookup'][
                self.sequence['order'][next]
            ]['path']
        else:
            return None

    def page_redirect(self, page_name):
        if page_name not in self.sequence['order']:
            raise Exception('Unrecognised SaleboxCheckout page_name: %s' % page_name)

        # if basket empty, redirect to the pre-checkout page, e.g.
        # typically the shopping basket
        if len(self.data['basket']) == 0:
            return settings.SALEBOX['CHECKOUT']['PRE_URL']

        # if this page is not marked accessible, i.e. the user is
        # trying to jump steps in the process, redirect them to
        # the 'last known good' page
        if not self.is_page_accessible(page_name):
            return self.get_last_accessible_page()

        return None

    def is_page_accessible(self, page_name):
        try:
            return self.sequence['lookup'][page_name]['accessible']
        except:
            return False

    def save_to_store(self, user, gateway_code='default', sent_to_gateway=True, visible_id=None):
        data = self.get_raw_data()
        cs = CheckoutStore(
            visible_id=visible_id,
            user=user.id if user else None,
            gateway_code=gateway_code,
            status=20 if sent_to_gateway else 10,
            data=self.get_raw_data()
        )
        cs.save()
        return cs

    def set_basket(self, basket, request, reset_completed=True, reset_checkout=True):
        if reset_checkout:
            self._init_data()
        if reset_completed:
            self.data['completed'] = []

        # populate data object
        self.data['basket'] = basket.get_raw_data()['basket']
        self._write_session(request)

        # return url to redirect to
        return self.sequence['lookup'][self.sequence['order'][0]]['path']

    def set_completed(self, page_name, request):
        self.data['completed'].append(page_name)
        self._update_sequence()
        self._write_session(request)
        return self.get_next_page(page_name)

    def set_customer_email(self, email, request):
        self.data['customer']['email'] = email
        self._write_session(request)

    def set_invoice_address(self, required, address_or_id, meta, request):
        if address_or_id is None:
            self.data['invoice_address']['address'] = None
        else:
            if isinstance(address_or_id, int):
                self.data['invoice_address']['address'] = model_to_dict(UserAddress.objects.get(id=address_or_id))
                self.data['invoice_address']['id'] = address_or_id
            else:
                self.data['invoice_address']['address'] = {
                    'default': address_or_id.default,
                    'address_group': address_or_id.address_group,
                    'full_name': address_or_id.full_name,
                    'address_1': address_or_id.address_1,
                    'address_2': address_or_id.address_2,
                    'address_3': address_or_id.address_3,
                    'address_4': address_or_id.address_4,
                    'address_5': address_or_id.address_5,
                    'country_state': model_to_dict(address_or_id.country_state) if address_or_id.country_state else None,
                    'country': model_to_dict(address_or_id.country) if address_or_id.country else None,
                    'id': None,
                    'postcode': address_or_id.postcode,
                    'phone_1': address_or_id.phone_1,
                    'phone_2': address_or_id.phone_2,
                    'email': address_or_id.email,
                    'string_1': address_or_id.string_1,
                    'string_2': address_or_id.string_2,
                    'tax_id': address_or_id.tax_id
                }
                self.data['invoice_address']['id'] = None

        self.data['invoice_address']['meta'] = meta
        self.data['invoice_address']['required'] = required
        self._write_session(request)

    def set_payment_method(self, id, meta, request):
        self.data['payment_method'] = {
            'id': int(id),
            'meta': meta
        }
        self._write_session(request)

    def set_shipping_address(self, required, address_or_id, meta, request):
        if isinstance(address_or_id, int):
            self.data['shipping_address']['address'] = model_to_dict(UserAddress.objects.get(id=address_or_id))
            self.data['shipping_address']['id'] = address_or_id
        else:
            self.data['shipping_address']['address'] = {
                'default': address_or_id.default,
                'address_group': address_or_id.address_group,
                'full_name': address_or_id.full_name,
                'address_1': address_or_id.address_1,
                'address_2': address_or_id.address_2,
                'address_3': address_or_id.address_3,
                'address_4': address_or_id.address_4,
                'address_5': address_or_id.address_5,
                'country_state': model_to_dict(address_or_id.country_state) if address_or_id.country_state else None,
                'country': model_to_dict(address_or_id.country) if address_or_id.country else None,
                'id': None,
                'postcode': address_or_id.postcode,
                'phone_1': address_or_id.phone_1,
                'phone_2': address_or_id.phone_2,
                'email': address_or_id.email,
                'string_1': address_or_id.string_1,
                'string_2': address_or_id.string_2,
                'tax_id': address_or_id.tax_id
            }
            self.data['shipping_address']['id'] = None

        self.data['shipping_address']['meta'] = meta
        self.data['shipping_address']['required'] = required
        self._write_session(request)

    def set_shipping_method(self, id, code, label, price, meta, request):
        self.data['shipping_method'] = {
            'code': code,
            'id': id,
            'label': label,
            'price': price,
            'meta': meta
        }
        self._write_session(request)

    def _init_data(self):
        self.data = {
            'basket': {},
            'completed': [],
            'customer': {
                'email': None
            },
            'data': {},
            'invoice_address': {
                'address': None,
                'id': None,
                'meta': None,
                'required': None,
            },
            'last_seen': int(time.time()),
            'payment_method': {
                'id': None,
                'meta': None,
            },
            'shipping_address': {
                'address': None,
                'id': None,
                'meta': None,
                'required': None,
            },
            'shipping_method': {
                'id': None,
                'label': None,
                'meta': None,
                'price': None,
            }
        }

    def _init_sequence(self):
        self.sequence = {
            'order': [],
            'lookup': {}
        }

        # use language url prefix? e.g. /en/checkout/address
        url_prefix = ''
        if settings.SALEBOX['CHECKOUT'].get('URL_LANGUAGE_PREFIX', False):
            url_prefix = '/%s' % get_language()

        for i, s in enumerate(settings.SALEBOX['CHECKOUT']['SEQUENCE']):
            self.sequence['order'].append(s[0])
            self.sequence['lookup'][s[0]] = {
                'label': s[2] if len(s) == 3 else s[0],
                'path': '%s%s' % (url_prefix, s[1]),
                'position': i,
                'complete': False,
                'accessible': i == 0
            }

    def _init_session(self, request):
        self._init_data()

        # attempt to import data from the session
        request.session.setdefault('saleboxcheckout', None)
        if request.session['saleboxcheckout'] is not None:
            tmp = request.session['saleboxcheckout']
            if int(time.time()) - tmp['last_seen'] < 60 * 60:  # 1 hr
                self.data = request.session['saleboxcheckout']

        # update data
        self._update_sequence()
        self.data['last_seen'] = int(time.time())

    def _update_sequence(self):
        for i, o in enumerate(self.sequence['order']):
            if o in self.data['completed']:
                self.sequence['lookup'][o]['complete'] = True
                self.sequence['lookup'][o]['accessible'] = True
                try:
                    self.sequence['lookup'][
                        self.sequence['order'][i + 1]
                    ]['accessible'] = True
                except:
                    pass

    def _write_session(self, request):
        if len(self.data['basket']) == 0:
            request.session['saleboxcheckout'] = None
        else:
            request.session['saleboxcheckout'] = self.data



from django.shortcuts import redirect
from django.views.generic import TemplateView

from saleboxdjango.lib.basket import SaleboxBasket
from saleboxdjango.lib.checkout import SaleboxCheckout


class SaleboxAccountBasketView(TemplateView):
    template_name = 'salebox/account/basket.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        sb = SaleboxBasket(request)
        context['data'] = sb.get_raw_data()

        sc = SaleboxCheckout(request)
        context['checkout'] = {
            'nav': sc.get_checkout_nav()
        }

        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        sb = SaleboxBasket(request)
        sc = SaleboxCheckout(request)
        return redirect(sc.set_basket(sb, request))



from django.conf import settings
from django.utils.translation import get_language

from python2c2p.redirectapi import twoctwop_redirectapi

from saleboxdjango.lib.basket import SaleboxBasket
from saleboxdjango.views.checkout.gateway import SaleboxCheckoutGatewayView


class SaleboxProviders2C2PGatewayView(SaleboxCheckoutGatewayView):
    gateway_code = '2c2p'

    def gateway(self, request, *args, **kwargs):
        context = self.get_context_data()
        store = self.sc.save_to_store(request.user, gateway_code=self.gateway_code)

        # override start
        data = self.sc.get_raw_data()

        # get the html to POST to the gatewy
        context['html'] = self.gateway_init(data, request, store)

        # reset basket  TODO - should the basket be reset on first callback?
        basket = SaleboxBasket(request)
        basket.reset_basket(request)

        # render
        return self.render_to_response(context)

    def gateway_init(self, data, request, store):
        """
        Optionally override this method if there are different
        payment parameters you wish to send
        """

        tctp = twoctwop_redirectapi(
            settings.TWOCTWOP['MERCHANT_ID'],
            settings.TWOCTWOP['SECRET_KEY'],
            settings.TWOCTWOP['GATEWAY_URL'],
        )
        tctp.set_value('payment_option', 'A')

        # total price
        total_price = data['basket']['sale_price'] + data['shipping_method']['price']

        # default values
        tctp.set_value('amount', total_price)
        tctp.set_value('currency', 'THB')
        tctp.set_value('customer_email', request.user.email)
        tctp.set_value('default_lang', get_language())
        tctp.set_value('order_id', store.visible_id)
        tctp.set_value('payment_description', settings.TWOCTWOP['PAYMENT_DESCRIPTION'])
        tctp.set_value('result_url_1', settings.TWOCTWOP['CALLBACK_URL_FRONTEND'],)
        tctp.set_value('result_url_2', settings.TWOCTWOP['CALLBACK_URL_BACKEND'],)
        tctp.set_value('user_defined_1', store.uuid)

        return tctp.request()



# Generated by Django 2.1.8 on 2019-08-29 07:07

import datetime
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0003_auto_20190827_0809'),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='salebox_transactionhistory_count',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='member',
            name='salebox_transactionhistory_data',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='member',
            name='salebox_transactionhistory_request_dt',
            field=models.DateTimeField(blank=True, default=datetime.datetime.now, null=True),
        ),
    ]



from django import forms

from saleboxdjango.lib.basket import SaleboxBasket
from .base import SaleboxCheckoutBaseView


class SaleboxCheckoutGatewayForm(forms.Form):
    pass


class SaleboxCheckoutGatewayView(SaleboxCheckoutBaseView):
    http_method_names = ['get', 'post']

    checkout_step = 'gateway'
    form_class = SaleboxCheckoutGatewayForm
    template_name = 'salebox/checkout/gateway.html'

    def gateway(self, request, *args, **kwargs):
        context = self.get_context_data()
        store = self.sc.save_to_store(request.user)

        # override this middle section relevant to your needs
        #
        #
        #
        context['html'] = '<form><!-- PUT YOUR CODE HERE --></form>'

        # reset the shopping basket
        basket = SaleboxBasket(request)
        basket.reset_basket(request)

        # render the gateway redirect
        return self.render_to_response(context)

    def get(self, request, *args, **kwargs):
        return self.gateway(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.gateway(request, *args, **kwargs)



import datetime
import re

from django.conf import settings
from django.contrib.auth import logout
from django.shortcuts import redirect
from django.utils.translation import LANGUAGE_SESSION_KEY

from saleboxdjango.lib.basket import SaleboxBasket


class SaleboxMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # kick out inactive (i.e. banned) users
        if request.user.is_authenticated and not request.user.is_active:
            request.session['saleboxbasket'] = None
            logout(request)
            return redirect('/')

        # init shopping basket
        sb = SaleboxBasket(request)

        # set product_list_order
        request.session.setdefault(
            'product_list_order',
            settings.SALEBOX['SESSION']['DEFAULT_PRODUCT_LIST_ORDER']
        )
        if 'product_list_order' in request.GET:
            valid_orders = [
                'bestseller_low_to_high',
                'bestseller_high_to_low',
                'price_low_to_high',
                'price_high_to_low',
                'rating_high_to_low',
                'rating_low_to_high',
            ]
            if request.GET['product_list_order'] in valid_orders:
                request.session['product_list_order'] = request.GET['product_list_order']
                if re.search(r'\d+\/$', request.path):
                    return redirect(re.sub(r'\d+\/$', '', request.path))

        # create response
        response = self.get_response(request)
        if sb.get_cookie_action(request) == 'add':
            response.set_cookie(
                'psessionid',
                value=request.session.session_key,
                max_age=60 * 60 * 24 * 365
            )
        elif sb.get_cookie_action(request) == 'remove':
            response.delete_cookie('psessionid')
        return response


"""
This middleware sets whatever language is in the URL, e.g. /en/about-us = 'en'
and stores it in the session[LANGUAGE_SESSION_KEY]. This means django
can 'remember' the language of the non-language specific URLs, e.g. /basket/
"""
class SaleboxI18NSessionStoreMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        available_languages = [l[0].lower() for l in settings.LANGUAGES]
        key = getattr(settings, 'LANGUAGE_SESSION_KEY', LANGUAGE_SESSION_KEY)
        prev_language = request.session.get(key, None)
        curr_language = prev_language

        # set a default language if none exists
        if curr_language is None:
            curr_language = available_languages[0]

        # attempt to set the language from the path
        try:
            prefix = request.path.lower().strip('/').split('/')[0]
            if prefix in available_languages:
                curr_language = prefix
        except:
            pass

        # update the session if applicable
        if prev_language != curr_language:
            request.session[key] = curr_language

        # create response
        return self.get_response(request)



from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView

from saleboxdjango.forms import SaleboxAddressAddForm
from saleboxdjango.lib.address import SaleboxAddress


class SaleboxAccountAddressView(TemplateView):
    initial = {}
    template_name = 'salebox/account/address.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.sa = SaleboxAddress(request.user)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        self.form = SaleboxAddressAddForm(initial=self.initial)
        return self.output(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.form = SaleboxAddressAddForm(
            request.POST,
            initial=self.initial
        )
        if self.form.is_valid():
            address = self.sa.add(self.form.cleaned_data)

            # redirect to self to prevent refresh
            return redirect(request.get_full_path())

        return self.output(request, *args, **kwargs)

    def output(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        context['addresses'] = self.sa.get()
        context['address_form'] = self.form
        context['address_extras'] = self.sa.form_extras(
            country_id=self.form['country'].value()
        )

        return self.render_to_response(context)



from django import forms

from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutPaymentMethodForm(forms.Form):
    payment_method = forms.ChoiceField(choices=(
        (1, 'Credit or Debit Card'),
        (2, 'E-Wallet'),
    ))


class SaleboxCheckoutPaymentMethodView(SaleboxCheckoutBaseView):
    checkout_step = 'payment_method'
    form_class = SaleboxCheckoutPaymentMethodForm
    template_name = 'salebox/checkout/payment_method.html'

    def form_valid_pre_redirect(self, form):
        id = form.cleaned_data['payment_method']
        self.sc.set_payment_method(id, None, self.request)
        return True



from saleboxdjango.views.address.base import SaleboxAddressView
from saleboxdjango.forms import SaleboxAddressAddForm

from saleboxdjango.models import Country, CountryState, UserAddress


class SaleboxAddressAddView(SaleboxAddressView):
    action = 'address-add'
    form = SaleboxAddressAddForm

    def form_valid(self, request):
        try:
            country = None
            if self.form.cleaned_data['country'] is not None:
                country = Country \
                            .objects \
                            .filter(id=self.form.cleaned_data['country']) \
                            .first()

            state = None
            if self.form.cleaned_data['country_state'] is not None:
                state = CountryState \
                            .objects \
                            .filter(id=self.form.cleaned_data['country_state']) \
                            .first()

            ua = UserAddress(
                user=request.user,
                default=self.form.cleaned_data['set_default'],
                address_group=self.form.cleaned_data['address_group'] or 'default',
                full_name=self.form.cleaned_data['full_name'],
                address_1=self.form.cleaned_data['address_1'],
                address_2=self.form.cleaned_data['address_2'],
                address_3=self.form.cleaned_data['address_3'],
                address_4=self.form.cleaned_data['address_4'],
                address_5=self.form.cleaned_data['address_5'],
                country_state=state,
                country=country,
                postcode=self.form.cleaned_data['postcode']
            )
            ua.save()
            self.status = 'success'
        except:
            self.status = 'fail'



from opposablethumbs.opposablethumbs import OpposableThumbs

from django.conf import settings
from django.http import Http404


def salebox_image_view(request, imgtype, dir, id, hash, suffix):
    try:
        f = '%s/salebox/%s/%s.%s.%s' % (
            settings.MEDIA_ROOT,
            dir,
            id,
            hash,
            suffix
        )
        params = 'file=%s&%s' % (f, settings.SALEBOX['IMG'][imgtype])
        ot = OpposableThumbs(params)
        return ot.response()
    except:
        raise Http404()



from django import forms
from django.conf import settings
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

from saleboxdjango.models import LastUpdate


class SaleboxSyncForm(forms.Form):
    key = forms.CharField()
    license = forms.CharField()


@csrf_exempt
def SaleboxSyncView(request):
    success = False

    # is this a POST?
    if request.method == 'POST':
        # did it come from the server IP?
        if settings.SALEBOX['API']['IP'] == request.META['REMOTE_ADDR']:
            # did it send a valid key / license combo?
            form = SaleboxSyncForm(request.POST)
            if form.is_valid():
                if  settings.SALEBOX['API']['KEY'] == str(form.cleaned_data['key']) and \
                    settings.SALEBOX['API']['LICENSE'] == str(form.cleaned_data['license']):
                    # ok, we're good. reset the LastUpdate value
                    lu = LastUpdate \
                            .objects \
                            .filter(code='saleboxsync_pull_start') \
                            .filter(value__gt=0) \
                            .first()
                    if lu is not None:
                        lu.value = 0
                        lu.save()
                    return JsonResponse({ 'status': 'OK' })

    # fallback
    return JsonResponse({ 'status': 'Fail' })


from django import forms
from .base import SaleboxCheckoutBaseView


class SaleboxCheckoutReceiptForm(forms.Form):
    pass


class SaleboxCheckoutReceiptView(SaleboxCheckoutBaseView):
    checkout_step = 'receipt'
    form_class = SaleboxCheckoutReceiptForm
    template_name = 'salebox/checkout/receipt.html'



from saleboxdjango.views.base import SaleboxBaseView
from saleboxdjango.lib.basket import SaleboxBasket


class SaleboxBasketView(SaleboxBaseView):
    user_must_be_authenticated = False

    def init_class(self, request):
        self.sb = SaleboxBasket(request)



from math import modf

from django import template
from django.contrib.humanize.templatetags.humanize import intcomma
from django.utils.safestring import mark_safe


register = template.Library()

@register.filter(is_safe=True)
def sb_currency(value):
    if not isinstance(value, int):
        try:
            value = int(value)
        except:
            value = 0
    minor, major = modf(value / 100)
    return mark_safe('%s<span>.%s</span>' % \
        (intcomma(int(major)), '{:.2f}'.format(minor)[2:]))

@register.filter(is_safe=True)
def sb_currency_text(value):
    if not isinstance(value, int):
        try:
            value = int(value)
        except:
            value = 0
    minor, major = modf(value / 100)
    return '%s.%s' % (intcomma(int(major)), '{:.2f}'.format(minor)[2:])

@register.filter(is_safe=True)
def sb_trans(obj, field):
    return obj.get_trans(field)



import importlib
import time

from django.conf import settings
from django.template.loader import render_to_string

from saleboxdjango.models import BasketWishlist, ProductVariant


class SaleboxBasket:
    def __init__(self, request):
        self.cookie = None
        self.data = None

        # init a default value if the session variable is not set
        # load the basket into memory if it exists
        request.session.setdefault('saleboxbasket', None)
        if request.session['saleboxbasket'] is not None:
            self.data = request.session['saleboxbasket']

        # if the user is not logged in, store their sessionid in a cookie
        # so we can use it later to fetch their anonymous basket
        key = request.session.session_key
        if not request.user.is_authenticated:
            if key is None:
                # just logged out, clear any lingering data
                self.data = None
            else:
                try:
                    if request.COOKIES['psessionid'] != key:
                        self.cookie = 'add'
                except:
                    self.cookie = 'add'

        # if the user is logged in, see if their sessionid has changed.
        # if so, they've just logged in and we need to migrate their
        # anonymous basket to their user one
        if request.user.is_authenticated:
            try:
                if request.COOKIES['psessionid'] != key:
                    self._migrate_anonymous_basket(
                        request.COOKIES['psessionid'],
                        request.user
                    )
                    self.cookie = 'remove'
                    self.data = None
            except:
                pass

        # check the time hasn't expired. if it has, rebuild the
        # basket details. this is to keep shopping baskets in sync
        # for those users using more than one device
        if self.data is not None:
            if int(time.time()) - request.session['saleboxbasket']['created'] > 300:
                self.data = None

        # if no data exists, create it
        if self.data is None:
            self._init_basket(request)

    def get_cookie_action(self, request):
        return self.cookie

    def get_data(self, request, results, variant_id=None):
        results = results.split(',')
        o = {}

        # construct output
        if 'all' in results:
            results = [
                'basket_html_button',
                'basket_html_full',
                'basket_html_summary',
                'basket_loyalty',
                'basket_price',
                'basket_qty_total',
                'basket_qty_variant',
                'wishlist_html_button',
                'wishlist_html_full',
                'wishlist_html_summary',
            ]

        if 'basket_html_button' in results and variant_id:
            try:
                qty = self.data['basket']['lookup'][str(variant_id)]['qty']
            except:
                qty = 0

            try:
                stock_total = self.data['basket']['lookup'][str(variant_id)]['variant']['stock_total']
            except:
                # this is a hack to prevent the button showing the
                # 'out of stock' message. Its works though - if we're
                # reaching this code, the variant is in stock
                stock_total = 1

            try:
                preorder_flag = self.data['basket']['lookup'][str(variant_id)]['variant']['preorder_flag']
            except:
                preorder_flag = False

            o['basket_html_button'] = render_to_string(
                'salebox/basket/button_basket.html',
                {
                    'pv': {
                        'id': variant_id,
                        'basket_qty': qty,
                        'preorder_flag': preorder_flag,
                        'stock_total': stock_total
                    },
                    'request': request
                },
                request
            )

        if 'basket_html_full' in results:
            o['basket_html_full'] = render_to_string(
                'salebox/basket/basket_full.html',
                {
                    'data': self.data,
                    'request': request
                },
                request
            )

        if 'basket_html_summary' in results:
            o['basket_html_summary'] = render_to_string(
                'salebox/basket/basket_summary.html',
                {
                    'data': self.data,
                    'request': request
                },
                request
            )

        if 'basket_loyalty' in results:
            o['basket_loyalty'] = self.data['basket']['loyalty']

        if 'basket_price' in results:
            o['basket_price'] = {
                'orig_price': self.data['basket']['orig_price'],
                'sale_price': self.data['basket']['sale_price']
            }

        if 'basket_qty_total' in results:
            o['basket_qty_total'] = self.data['basket']['qty']

        if 'basket_qty_variant' in results and variant_id:
            o['basket_qty_variant'] = len(self.data['basket']['order'])

        if 'wishlist_html_button' in results and variant_id:
            in_wishlist = str(variant_id) in self.data['wishlist']['lookup']

            o['wishlist_html_button'] = render_to_string(
                'salebox/basket/button_wishlist.html',
                {
                    'pv': {
                        'id': variant_id,
                        'in_wishlist': in_wishlist
                    },
                    'request': request
                },
                request
            )

        if 'wishlist_html_full' in results:
            o['wishlist_html_full'] = render_to_string(
                'salebox/basket/wishlist_full.html', {
                    'data': self.data,
                    'request': request
                },
                request
            )

        if 'wishlist_html_summary' in results:
            o['wishlist_html_full'] = render_to_string(
                'salebox/basket/wishlist_summary.html', {
                    'data': self.data,
                    'request': request
                },
                request
            )

        return o

    def get_raw_data(self):
        return self.data

    def migrate_basket_wishlist(self, request, variant, to_basket):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        bwl = self._filter_basket_queryset(
            request,
            BasketWishlist
                .objects
                .filter(variant=variant)
                .exclude(basket_flag=to_basket)
        )

        for b in bwl:
            b.basket_flag = to_basket
            b.quantity = 1
            b.save()

        # re-populate basket
        self._init_basket(request)

    def reset_basket(self, request):
        # get items from db
        self._filter_basket_queryset(
            request,
            BasketWishlist.objects.filter(basket_flag=True)
        ).delete()
        self._init_basket(request)

    def update_basket(self, request, variant, qty, relative):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        # get existing items to update
        items = self._get_variants(request, variant)

        # update db
        if items['basket'] is None:
            if items['wishlist'] is None:
                # create basket item
                if int(qty) > 0:
                    self._add_variant(request, variant, qty, True)
            else:
                # wishlist item exists, convert to basket
                items['wishlist'].basket_flag = True
                items['wishlist'].quantity = int(qty)
                items['wishlist'].save()
        else:
            # update basket item
            if relative:
                items['basket'].quantity += int(qty)
            else:
                items['basket'].quantity = int(qty)
            if items['basket'].quantity > 0:
               items['basket'].save()
            else:
                items['basket'].delete()

            # delete wishlist item if exists
            if items['wishlist'] is not None:
                items['wishlist'].delete()

        # re-populate basket
        self._init_basket(request)

    def update_wishlist(self, request, variant, add):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        # get existing items to update
        items = self._get_variants(request, variant)

        # update db
        if add:
            if items['wishlist'] is None:
                # add to wishlist
                self._add_variant(request, variant, 1, False)
        else:
            if items['wishlist'] is not None:
                items['wishlist'].delete()

        # re-populate basket
        self._init_basket(request)

    def _add_variant(self, request, variant, qty, basket):
        bwl = BasketWishlist(
            variant=variant,
            quantity=int(qty),
            basket_flag=basket,
        )
        if request.user.is_authenticated:
            bwl.user = request.user
        else:
            bwl.session = request.session.session_key
        bwl.save()

    def _calculate_loyalty(self, request):
        # pre_calculate_loyalty
        self._call_external('PRE_CALCULATE_LOYALTY', request)

        for i in self.data['basket']['items']:
            if i['variant']['loyalty'] is not None:
                i['variant']['qty_loyalty'] = i['variant']['loyalty'] * i['qty']
                self.data['basket']['loyalty'] += i['variant']['loyalty'] * i['qty']

        # post_calculate_loyalty
        self._call_external('POST_CALCULATE_LOYALTY', request)

    def _calculate_price(self, request):
        # pre_calculate_price
        self._call_external('PRE_CALCULATE_PRICE', request)

        for i in self.data['basket']['items']:
            i['variant']['qty_price'] = i['variant']['price'] * i['qty']
            i['variant']['qty_sale_price'] = i['variant']['sale_price'] * i['qty']
            self.data['basket']['orig_price'] += i['variant']['price'] * i['qty']
            self.data['basket']['sale_price'] += i['variant']['sale_price'] * i['qty']
            if i['variant']['preorder_flag']:
                self.data['basket']['orig_price_preorder'] += i['variant']['price'] * i['qty']
                self.data['basket']['sale_price_preorder'] += i['variant']['sale_price'] * i['qty']
            else:
                self.data['basket']['orig_price_onshelf'] += i['variant']['price'] * i['qty']
                self.data['basket']['sale_price_onshelf'] += i['variant']['sale_price'] * i['qty']

        self.data['basket']['saving'] = self.data['basket']['orig_price'] - self.data['basket']['sale_price']

        # post_calculate_price
        self._call_external('POST_CALCULATE_PRICE', request)

    def _call_external(self, name, request):
        if settings.SALEBOX['CHECKOUT']:
            funcstr = settings.SALEBOX['CHECKOUT'].get(name)
            if funcstr:
                pkg, attr = funcstr.rsplit('.', 1)
                func = getattr(importlib.import_module(pkg), attr)
                self.data = func(self.data, request)

    def _filter_basket_queryset(self, request, qs):
        qs = qs.select_related(
            'variant',
            'variant__product',
            'variant__product__category',
        )
        qs = qs.order_by('created')

        if request.user.is_authenticated:
            return qs.filter(user=request.user) \
                     .filter(session__isnull=True)
        else:
            return qs.filter(user__isnull=True) \
                     .filter(session=request.session.session_key)

    def _get_raw_basket(self, request):
        while True:
            # get basket items
            basket = self._filter_basket_queryset(
                request,
                BasketWishlist.objects.filter(basket_flag=True)
            )

            # do not allow basket qty to exceed available qty
            for b in basket:
                if not b.variant.preorder_flag and b.variant.product.inventory_flag:
                    if b.quantity > 0 and b.quantity > b.variant.stock_total:
                        if b.variant.stock_total < 1:
                            b.quantity = 1
                            b.basket_flag = False
                        else:
                            b.quantity = b.variant.stock_total

                        b.save()

            # get items from db
            qs = self._filter_basket_queryset(
                request,
                BasketWishlist.objects.all()
            )

            # init duplicate lookups
            duplicate_basket = {}
            duplicate_wishlist = {}
            delete_ids = []

            # find duplicates
            for pos, q in enumerate(qs):
                vid = str(q.variant.id)

                if q.basket_flag:
                    # basket
                    if vid in duplicate_basket:
                        # merge into the first instance of this variant
                        qs[duplicate_basket[vid]].quantity += q.quantity
                        qs[duplicate_basket[vid]].save()

                        # mark the duplicate to be deleted
                        delete_ids.append(q.id)
                    else:
                        duplicate_basket[vid] = pos
                else:
                    # wishlist
                    if vid in duplicate_wishlist:
                        delete_ids.append(q.id)
                    else:
                        if q.quantity != 1:
                            q.quantity = 1
                            q.save()
                        duplicate_wishlist[vid] = pos

            # delete duplicates if they exist
            if len(delete_ids) > 0:
                BasketWishlist.objects.filter(id__in=delete_ids).delete()
            else:
                break

        return qs

    def _get_variants(self, request, variant):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        basket = self._filter_basket_queryset(
            request,
            BasketWishlist
                .objects
                .filter(basket_flag=True)
                .filter(variant=variant)
        )

        wishlist = self._filter_basket_queryset(
            request,
            BasketWishlist
                .objects
                .filter(basket_flag=False)
                .filter(variant=variant)
        )

        return {
            'basket': basket.first(),
            'wishlist': wishlist.first(),
        }

    def _init_basket(self, request):
        self.data = {
            'created': int(time.time()),
            'basket': {
                'qty': 0,
                'qty_onshelf': 0,
                'qty_preorder': 0,
                'orig_price': 0,
                'orig_price_onshelf': 0,
                'orig_price_preorder': 0,
                'sale_price': 0,
                'sale_price_onshelf': 0,
                'sale_price_preorder': 0,
                'loyalty': 0,
                'items': [],
                'order': [],
                'lookup': {}
            },
            'wishlist': {
                'qty': 0,
                'qty_onshelf': 0,
                'qty_preorder': 0,
                'order': [],
                'items': [],
                'lookup': {}
            }
        }

        # retrieve items from db
        qs = self._get_raw_basket(request)

        # populate data
        for q in qs:
            pv = {
                'default_image': q.variant.default_image,
                'ecommerce_description': q.variant.ecommerce_description,
                'id': q.variant.id,
                'image': q.variant.image,
                'int_1': q.variant.int_1,
                'int_2': q.variant.int_2,
                'int_3': q.variant.int_3,
                'int_4': q.variant.int_4,
                'local_image': q.variant.local_image,
                'loyalty': q.variant.loyalty_points,
                'member_discount_applicable': q.variant.member_discount_applicable,
                'name': q.variant.name,
                'plu': q.variant.plu,
                'preorder_flag': q.variant.preorder_flag,
                'price': q.variant.price,
                'product': {
                    'category': {
                        'id': q.variant.product.category.id,
                        'name': q.variant.product.category.name,
                        'short_name': q.variant.product.category.short_name,
                    },
                    'id': q.variant.product.id,
                    'image': q.variant.product.image,
                    'local_image': q.variant.product.local_image,
                    'name': q.variant.product.name,
                    'slug': q.variant.product.slug,
                    'sold_by': q.variant.product.sold_by,
                    'string_1': q.variant.product.string_1,
                    'string_2': q.variant.product.string_2,
                    'string_3': q.variant.product.string_3,
                    'string_4': q.variant.product.string_4,
                    'vat_applicable': q.variant.product.vat_applicable,
                },
                'rating_score': q.variant.rating_score,
                'rating_vote_count': q.variant.rating_vote_count,
                'sale_percent': q.variant.sale_percent,
                'sale_price': q.variant.sale_price,
                'shipping_weight': q.variant.shipping_weight,
                'size_uom': q.variant.size_uom,
                'sku': q.variant.sku,
                'slug': q.variant.slug,
                'stock_total': q.variant.stock_total,
                'string_1': q.variant.string_1,
                'string_2': q.variant.string_2,
                'string_3': q.variant.string_3,
                'string_4': q.variant.string_4,
            }

            if q.basket_flag:
                self.data['basket']['qty'] += q.quantity
                if q.variant.preorder_flag:
                    self.data['basket']['qty_preorder'] += q.quantity
                else:
                    self.data['basket']['qty_onshelf'] += q.quantity
                self.data['basket']['order'].append(str(q.variant.id))
                self.data['basket']['lookup'][str(q.variant.id)] = {
                    'qty': q.quantity,
                    'variant': pv
                }
            else:
                self.data['wishlist']['qty'] += q.quantity
                if q.variant.preorder_flag:
                    self.data['wishlist']['qty_preorder'] += q.quantity
                else:
                    self.data['wishlist']['qty_onshelf'] += q.quantity
                self.data['wishlist']['order'].append(str(q.variant.id))
                self.data['wishlist']['lookup'][str(q.variant.id)] = pv

        # populate the items list
        for i in self.data['basket']['order']:
            self.data['basket']['items'].append(self.data['basket']['lookup'][str(i)])
        for i in self.data['wishlist']['order']:
            self.data['wishlist']['items'].append(self.data['wishlist']['lookup'][str(i)])

        # calculate basket value + loyalty points
        self._calculate_price(request)
        self._calculate_loyalty(request)

        # save to session
        request.session['saleboxbasket'] = self.data

    def _migrate_anonymous_basket(self, key, user):
        # get all basket items from previously anonymous visitor
        basket = BasketWishlist \
                    .objects \
                    .filter(user__isnull=True) \
                    .filter(session=key) \

        for b in basket:
            b.user = user
            b.session = None
            b.save()



from django.conf import settings
from django.shortcuts import redirect
from django.views.generic import FormView

from saleboxdjango.lib.checkout import SaleboxCheckout


class SaleboxCheckoutBaseView(FormView):
    checkout_step = None

    def get_conf(self, name, default):
        return settings.SALEBOX['CHECKOUT'].get(name, default)

    def dispatch(self, request, *args, **kwargs):
        # friendlier error messages
        if self.checkout_step is None:
            raise Exception('You need to define a checkout_step')
        if self.form_class is None:
            raise Exception('You need to define a form_class')

        # get defaults
        user_must_be_authenticated = self.get_conf(
            'CHECKOUT_USER_MUST_BE_AUTHENTICATED',
            True
        )
        user_not_authenticated_redirect = self.get_conf(
            'CHECKOUT_USER_NOT_AUTHENTICATED_REDIRECT',
            '/'
        )

        # check logged in
        if user_must_be_authenticated and not request.user.is_authenticated:
            return redirect(user_not_authenticated_redirect)

        # get checkout object
        self.sc = SaleboxCheckout(request)
        r = self.sc.page_redirect(self.checkout_step)
        if r is not None:
            return redirect(r)

        # store the request in the object
        self.request = request

        # default dispatch action
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        valid = self.form_valid_pre_redirect(form)
        if not valid:
            return self.form_invalid(form)

        # set as complete and redirect to the next step
        r = self.sc.set_completed(self.checkout_step, self.request)
        if r is None:
            raise Exception('There is no next checkout step to redirect to...')
        else:
            return redirect(r)

    def form_valid_pre_redirect(self, form):
        # add you own code here
        # self.request is available
        # return True if form handled correctly
        # return False to re-show the page as form_invalid()
        #
        return True

    def get_additional_context_data(self, context):
        # add your custom code here, e.g...
        #     context['foo'] = 'bar'
        # ...then just return it
        # self.request is available
        #
        return context

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        data = self.sc.get_raw_data()

        total_price = data['basket']['sale_price']
        if data['shipping_method']['id'] is not None:
            total_price += data['shipping_method']['price']

        context['checkout'] = {
            'data': data,
            'nav': self.sc.get_checkout_nav(self.checkout_step),
            'step': self.checkout_step,
            'total_price': total_price
        }
        return self.get_additional_context_data(context)



# Generated by Django 2.1.8 on 2019-10-06 04:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0009_auto_20191005_1501'),
    ]

    operations = [
        migrations.AddField(
            model_name='checkoutstore',
            name='payment_received',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from saleboxdjango.views.address.base import SaleboxAddressView
from saleboxdjango.forms import SaleboxAddressIDForm


class SaleboxAddressRemoveView(SaleboxAddressView):
    action = 'address-remove'
    form = SaleboxAddressIDForm

    def form_valid(self, request):
        try:
            self.sa.set_inactive(self.form.cleaned_data['id'])
            self.status = 'success'
        except:
            self.status = 'fail'



from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView

from saleboxdjango.lib.address import SaleboxAddress


class SaleboxAccountHomeView(TemplateView):
    template_name = 'salebox/account/home.html'

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        sa = SaleboxAddress(request.user)
        context['delivery_address_count'] = sa.get_count()

        return self.render_to_response(context)



# Generated by Django 2.1.8 on 2019-10-05 05:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0006_translation'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='country',
            options={'ordering': ['code'], 'verbose_name_plural': 'Countries'},
        ),
        migrations.AlterModelOptions(
            name='countrystate',
            options={'ordering': ['full_code'], 'verbose_name_plural': 'Country States'},
        ),
        migrations.RenameField(
            model_name='country',
            old_name='code_2',
            new_name='code',
        ),
        migrations.RenameField(
            model_name='countrystate',
            old_name='code_2',
            new_name='code',
        ),
        migrations.RemoveField(
            model_name='country',
            name='code_3',
        ),
        migrations.RemoveField(
            model_name='country',
            name='name',
        ),
        migrations.RemoveField(
            model_name='countrystate',
            name='name',
        ),
        migrations.AddField(
            model_name='countrystate',
            name='full_code',
            field=models.CharField(blank=True, max_length=8, null=True),
        ),
    ]
